<?php

// ===================================================
//               X.able PHP XML module
//           (C)2017 maciej@maciejnowak.com
// ===================================================
// compatibility: php6+
// build: 20201019
// ===================================================

    // ====== array to XML FILE CONTENT / begin ======
    function XmlFileContent($xml) {
    // -------------------------------------------
    // $xml = <array> Xable XML array
    // -------------------------------------------
    // RETURNS: <string> Ready to save xml file content
    // -------------------------------------------
        if(is_array($xml) && count($xml) > 0) {
            $xml = arrayToXml($xml, 1);
            return "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n<xable>\n".$xml."</xable>";
        };
    }
    // ====== array to XML FILE CONTENT / end ======

    // ====== GET FILENAME / begin ======
    function getFilename($file_path) {
    // -------------------------------------------
    // $path = <string> Xable document path
    // -------------------------------------------
    // RETURNS: <string> Filename, also for multiple extensions cases
    // -------------------------------------------
        $file_path = path($file_path, "filename");
        while(path($file_path, "extension") != "") {
            $file_path = path($file_path, "filename");
        }
        return $file_path;
    };
    // ====== GET FILENAME / end ======

    // ====== XML document EXISTS test / begin ======
    function xmlExists($path) {
    // -------------------------------------------
    // $path = <string> Xable document path
    // -------------------------------------------
    // RETURNS: <array> Document path & active mode ("published", "draft", "unpublished"),
    // eg: [ ../pages/file.xml, "published" ],
    // or <false> if file not exist
    // -------------------------------------------
        $filename = path($path, "filename");
        if(path($filename, "extension") == "xml") { $filename = path($path, "filename"); }
        $folder = path($path, "dirname");
        $ext_list = [ "xml", "xml.draft" ];
        $exists = [];
        foreach($ext_list as $ext) {
            if(file_exists("$folder/$filename.$ext")) { $exists[] = $ext; }
        };
        if(in_array("xml.draft", $exists) && in_array("xml", $exists)) {
            return [ "$folder/$filename.xml.draft", "draft" ];
        }
        elseif(!in_array("xml.draft", $exists) && in_array("xml", $exists)) {
            return [ "$folder/$filename.xml", "published" ];
        }
        elseif(in_array("xml.draft", $exists) && !in_array("xml", $exists)) {
            return [ "$folder/$filename.xml.draft", "unpublished" ];
        }
        else {
            return false;
        }
    };
    // ====== XML document EXISTS test / end ======

    // ====== get XML all LANGUAGES TITLES / begin ======
    function xmlLanguageTitles($xml_item) {
    // -------------------------------------------
    // $xml_item = <array> Item xable array
    // -------------------------------------------
    // RETURNS: <string> Titles in all languages
    // -------------------------------------------
        $item_title = "";
        foreach(array_keys($xml_item["title"][0]["text"][0]) as $language) {
            $title = $xml_item["title"][0]["text"][0][$language][0];
            $item_title = $item_title."<span class='lang_title $language'>$title</span>";
        }
        return $item_title;
    };
    // ====== get XML all LANGUAGES TITLES / end ======

    // ====== get XML document TITLE / begin ======
    function xmlTitle($path, $language = false, $draft = false) {
    // -------------------------------------------
    // $path = <string> Xable document path
    // $all_langs = <string> for specified language Title, "*": for all languages, or <false> for Admin Language
    // -------------------------------------------
    // RETURNS: <string> Document title based on header-title or filename
    // -------------------------------------------
        if(!is_string($language)) { $language = $_SESSION["admin_lang"]; };
        // Load draft support
        if($draft && file_exists("$path.$draft")) {
            $path = "$path.$draft";
        };
        // Variables
        $xml = loadXml($path, "draft");
        $headers = [ "header", "user", "person" ];
        $title = false;
        // Find title section
        foreach($headers as $section_name) {
            if(!$title && is_array($xml[$section_name])) {
                if($xml[$section_name][0]["title"]) {
                    // All languages
                    if($language == "*") {
                        $title = xmlLanguageTitles($xml[$section_name][0]);
                    }
                    // Single language
                    else {
                        $title = readXml($xml, "header title", $language);
                    };
                }
                else {
                    // User name
                    $key = array_shift(array_keys($xml["user"][0]));
                    $title = readXml($xml, "user $key", $language);
                }
            }
        };
        // No title -> Filename
        if(!is_string($title) || trim($title) == "") {
            $title = $path;
            while(path($title, "extension") != "") {
                $title = ucwords(path($title, "filename"));
            }
        }
        return $title;
    };
    // ====== get XML document TITLE / end ======

// ===================================================
//             Array <-> XML <-> JSON
// ===================================================

    // ====== load XML data to array / begin ======
    function loadXml($file, $preview = false, $force_preview = false) {
    // ----------------------------------------
    // $file = <string> File path
    // $preview = <string> Preview file extension or <false/undefined> for no preview
    // $force_preview = <boolean> Force to load preview, if any (needed for CMS)
    // ----------------------------------------
    // RETURNS: <array> XML data transformed into an array ([key] => ([0] => "val_1", [1] => "val_2)...)
    // ----------------------------------------
        if(is_string($preview) && $preview != "" && ($force_preview || $_SESSION['preview_mode'] == $preview) && file_exists("$file.$preview")) {
            $file = "$file.$preview";
            if(!$force_preview) { $_SESSION['preview_loaded'] = true; };
        };
        $xml = join(" ", array_map("trim", file($file)));
        $xml = xmlToArray($xml);
        if(array_shift(array_keys($xml)) == "xable") { $xml = $xml['xable'][0]; };
        return $xml;
    }; // ====== load XML data to array / end ======

    // ====== save array to XML file / begin ======
    function saveXml($file, $array) {
    // ----------------------------------------
    // $file = <string> File path
    // $array = <array> Data array, xml compatibile: ([key] => ([0] => "val_1", [1] => "val_2)...)
    // ----------------------------------------
    // Saves input data in XML format file
    // ----------------------------------------
        if(file_exists($file)) { rename($file, "$file.bak"); };
        $xml = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
        $xml .= arrayToXml($array, 0);
        file_put_contents($file, $xml);
        return $xml;
    }; // ====== save array to XML file / end ======

    // ====== XML to Array / begin ======
    function xmlToArray($xml) {
    // ----------------------------------------
    // $xml = <string> XML contents
    // ----------------------------------------
    // RETURNS: <array> XML data transformed into an array ([key] => ([0] => "val_1", [1] => "val_2)...)
    // ----------------------------------------
        $cmt_tags = array("<!--", "-->");
        unset($array);
        $n = 0;
        $text = "";         // text content
        $cmt = false;       // comments flag
        $tag = false;       // part of tag name
        $key = false;       // tag name = array key
        $sub = array();     // sub tags (will be ignored)
        while($n < strlen($xml)) {
            //echo "INPUT: $xml<br>";
            $ch = substr($xml, $n, 1);
            if($n < (strlen($xml) - 1)) { $next = substr($xml, ($n +1), 1); } else { $next = ""; };
            // ====== Ignore comments ======
            if($cmt != false) {
                if(substr($xml, $n, strlen($cmt_tags[1])) == $cmt_tags[1]) {
                    $cmt = false;
                    $n = $n + strlen($cmt_tags[1]) - 1;
                };
            }
            elseif(substr($xml, $n, strlen($cmt_tags[0])) == $cmt_tags[0]) { $cmt = true; }
            // ====== XML analyse ======
            elseif($ch == "<" && !is_numeric($next) && !in_array($next, array("", " ", "!", "?", ".", ",", ";")))  { // tag begin
                if(is_string($tag)) { $text .= "<".$tag; }; // previous "<" character was non-tag
                $tag = "";
            }
            elseif(is_string($tag) && $ch == ">") { // tag end
                if($key == false) { // new begin tag -> key
                    $key = $tag;
                    list($tag, $text) = array(false, "");
                }
                elseif($tag == "/".end($sub)) { // closing the last of sub tags
                    array_pop($sub);
                    $text .= "<".$tag.">";
                    $tag = false;
                }                    
                elseif($tag == "/".$key) { // matched key/tag begin & end                    
                    $text = trim($text);
                    $dat = $array[$key];
                    if(($sub = xmlToArray($text)) != false) { // recurence -> sub tags arrays
                        $dat[] = $sub;
                    }
                    else { // no sub tags -> string value
                        $dat[] = $text;
                    };
                    $array[$key] = $dat;
                    list($key, $tag, $text, $sub) = array(false, false, "", array()); // reset variables after adding complete tag data
                }
                else {
                    if(substr($tag, 0, 1) != "/") { $sub[] = $tag; }; // opening of the sub tag
                    $text .= "<".$tag.">";
                    $tag = false;
                };
            }
            elseif(is_string($tag)) { $tag .= $ch; }
            else {$text .= $ch; };
            $n++;
        };
        return $array;
    }; // ====== XML to Array / end ======

    // ====== Array to XML / begin ======
    function arrayToXml($array, $depth = 0) {
    // ----------------------------------------
    // $array = <array> Data array, xml compatibile: ([key] => ([0] => "val_1", [1] => "val_2)...)
    // $depth = <integer> array/xml level depth (optional)
    // ----------------------------------------
    // RETURNS: <string> data in XML string
    // ----------------------------------------
        if(!is_numeric($depth)) { $depth = 0; };
        $xml = "";
        foreach(array_keys($array) as $key) {
            $n = 0;
            $marg = "";
            while($n++ < $depth) { $marg .= "\t"; };
            $sub = $array[$key];
            foreach($sub as $val) {
                $xml .= "$marg<$key>";
                if(is_array($val)) {
                    $xml .= "\n";
                    $xml .= arrayToXml($val, ($depth + 1));
                    $xml .= "$marg</$key>\n";
                }
                else { $xml .= "$val</$key>\n"; };
            };
        };
        return $xml;
    }; // ====== Array to XML / end ======

	// ====== Array To JSON / begin ======
    function arrayToJson($array, $title = false, $depth = 0) {
    // -----------------------------------------------
	// $array = <array> Input array
    // $title = <string> Array/subarray title
    // $depth = <integer> Left indent depth
    // -----------------------------------------------
    // RETURNS:
    // <string> Input array in JSON format
    // -----------------------------------------------
        $json = [];
        $tab = "";
        for($i = 0; $i < $depth; $i++) { $tab = $tab."\t"; }
        if(is_int(array_shift(array_keys($array)))) { $brackets = [ "[", "]" ]; }
        else { $brackets = [ "{", "}" ]; };
        
        if(!$array) {
            $json[] = $tab."\"".$title."\": false,";
        }
        
        else if($title) { $json[] = $tab."\"".$title."\": ".$brackets[0]; };
        foreach(array_keys($array) as $key) {
            $val = $array[$key];
            if(is_string($val)) { $val = str_replace("\"", "\\\"", $val); } // quotes
            
            if(is_array($val)) {
                if(is_int($key)) { $key = false; };
                $json[] = arrayToJson($val, $key, ($depth + 1));
            }
            elseif($brackets[0] == "[") {
                $json[] = "$tab\t\"$val\",";
            }
            elseif(!$val) {
                $json[] = "$tab\t\"$key\": false,";
            }
            else {
                $json[] = "$tab\t\"$key\": \"$val\",";
            }
        }
        $last = array_pop(array_keys($json));
        $json[$last] = substr($json[$last], 0, (strlen($json[ $last ]) - 1));
        if($title && $array) { $json[] = $tab.$brackets[1]; }
        $json = join("\n", $json);
        if(!$title) { $json = "$tab{\n".$json."\n$tab}"; };
        if($depth > 0) { $json = $json.","; };
        return $json;
    }; // ====== Array To JSON / end ======

    // ====== Simplify XML array / begin ======
    function simplifyXmlArray($xml_array, $key = "href") {
    // -------------------------------------------
    // $xml_array = <array> XML format array
    // $key = <string> Main key name
    // -------------------------------------------
    // RETURNS: <array> Standard php associative array
    // -------------------------------------------
        $array = [];
        foreach($xml_array as $item) {
            if(isset($item[$key])) {
                $key_value = $item[$key];
                foreach(array_keys($item) as $sub_key) {
                    if($sub_key != $key) {
                        $sub_value = $item[$sub_key];
                        if(is_array($sub_value)) {
                            $sub_value = simplifyXmlArray($sub_value, $key);
                        }
                        $array[$key_value][$sub_key] = $sub_value;
                    }
                }
            }
        }
        return $array;
    };
    // ====== Simplify XML array / end ======

    // ====== Make XML array / begin ======
    function makeXmlArray($array, $key = "href") {
    // -------------------------------------------
    // $array = <array> associative array
    // $key = <string> Main key name
    // -------------------------------------------
    // RETURNS: <array> XML format array
    // -------------------------------------------
        $xml_array = [];
        foreach(array_keys($array) as $key_value) {
            $xml = [ $key => $key_value ];
            $item = $array[$key_value];
            foreach(array_keys($item) as $sub_key) {
                $sub_value = $item[$sub_key];
                if(is_array($sub_value)) {
                    $sub_value = makeXmlArray($sub_value, $key);
                }
                $xml[$sub_key] = $sub_value;
            }
            $xml_array[] = $xml;
        }
        return $xml_array;
    };
    // ====== Make XML array / end ======

    // ====== JSON to XML array conversion / begin ======
    function jsonToXmlArray($json, $key = "href") {
    // -------------------------------------------
    // $json = <string> JSON data
    // $key = <string> Main key name
    // -------------------------------------------
    // RETURNS: <array> XML format array
    // -------------------------------------------
        $array = json_decode($json, true);
        $xml_array = makeXmlArray($array, $key);
        return $xml_array;
    };
    // ====== JSON to XML array conversion / end ======

    // ====== XML array to JSON conversion / begin ======
    function xmlArrayToJson($xml_array, $key = "href") {
    // -------------------------------------------
    // $xml_array = <array> XML format array
    // $key = <string> Main key name
    // -------------------------------------------
    // RETURNS: <string> JSON data
    // -------------------------------------------
        $array = simplifyXmlArray($xml_array, $key);
        $json = arrayToJson($array);
        //$json = json_encode($array,  JSON_PRETTY_PRINT | JSON_UNESCAPED_UNICODE | JSON_ERROR_UTF8);
        return $json;
    };
    // ====== XML array to JSON conversion / end ======

// ===================================================
//                    Read XML data
// ===================================================

    // ====== getCode text / begin ======
    function getCode($code) {
    // ----------------------------------------
    // $code = <string> code from XML
    // ----------------------------------------
    // RETURNS = <string> code restored to original source code
    // ----------------------------------------
        $code = str_replace("&lt;", "<", $code);
        $code = str_replace("&gt;", ">", $code);
        $code = str_replace("[br]", "\n", $code);
        return $code;
    }
    // ====== getCode text / begin / end ======

    // ====== convert string to ID / begin ======
    function getId($string) {
    // ----------------------------------------
    // $string = <string>
    // ----------------------------------------
    // RETURNS = <string> Lowercase without spaces & special characters
    // ----------------------------------------
        $string = strtolower(killPl($string));
        $string = preg_replace("/ |\n|\t/", "_", $string);
        $string = preg_replace("/\||=|\^|\/|\\\|\*|'|\"|%|\\$|@|#|&|\\.|,|:|;|\?|!/", "", $string);
        $string = preg_replace("/<|>|\[|\]|\(|\)|{|}/", "", $string);
        $string = preg_replace("/___|__/", "_", $string);
        return $string;
    };
    // ====== convert string to ID / begin / end ======

    // ====== easy GET values from zable article / begin ======
    function getVal($article) {
    // ----------------------------------------
    // $article = <array> xable article array
    // ----------------------------------------
    // RETURNS = <array> All article sections values
    // ----------------------------------------
      $val = array();
      foreach(array_keys($article) as $key) {
          $val[$key] = readXml($article, $key);
      };
      return $val;
    };
    // ====== easy GET values from zable article / end ======

    // ====== easy READ XML data / begin ======
    function readXml($xml, $tags, $language = false) {
    // ----------------------------------------
    // $xml = <array> XML data loaded into an array with loadXml()
    // $tags = <string> TAGS list, divided with spaces, eg: "article section"
    // $language = <string> force language code, optional
    // ----------------------------------------
    // RETURNS: <string> data specified by tags or <array> if multiple data found
    // ----------------------------------------
        // Language
        if(!is_string($language)) { $language = $_SESSION['lang']; }; // By user session
        if(!is_string($language)) { $language = "pl"; } // Default
        // XML Array read
        $tags = explode(" ", $tags);
        foreach(array_keys($tags) as $n) {
            // ====== LOOP ======
            $tag = $tags[$n];
            if(($n + 1) < count($tags)) {
                $xml = $xml[$tag][0];
            }
            // ====== OUTPUT ======
            elseif(count($xml[$tag]) > 1) {
                // Multiple -> Array output
                return $xml[$tag];
            }
            else {
                // Unknown -> Guess output
                if(!is_array($xml[$tag][0]) && !is_string($xml[$tag][0])) {
                    if(is_string($xml['media'][0][$tag][0])) {
                        return $xml['media'][0][$tag][0];
                    }
                    elseif(is_array($xml['radio'][0][$tag])) {
                        return $xml['radio'][0][$tag];
                    }
                    elseif(is_array($xml['checkbox'][0][$tag])) {
                        return $xml['radio'][0][$tag];
                    };
                }
                // Single -> String output
                else {
                    $xml = $xml[$tag][0];
                    // Section data
                    if(is_array($xml['type']) && is_string($type = $xml['type'][0])) {
                        if($type == "text" || $type == "textarea") {
                            $text = $xml['text'][0][$language][0];
                            //if($xml['format'] == "") { return $text; } else { return BBCode($text); };
                            return $text;
                        }
                        elseif($type == "option") {
                            return $xml['selected'][0];
                        }
                        elseif($type == "media") {
                            return $xml[$type][0][ $xml['set'][0] ][0];
                        }
                        elseif($type == "button") {
                            return $xml['action'][0];
                        }
                        elseif($type == "table") {
                            return $xml[$type][0][$language];
                        }
                        else {
                            return $xml[$type][0];
                        };
                    }
                    // Specified options array
                    elseif(is_array($xml['option']) && ($tag == "radio" || $tag == "checkbox")) {
                        return ($xml['option']);
                    }
                    // Other data
                    else {
                        return $xml;
                    };
                };
            };
        };
    };
    // ====== easy READ XML data / begin ======

    // ====== TEST if MEDIA data is filled / begin ======
    function testMedia($val) {
    // ----------------------------------------
    // $val = <string> media content
    // ----------------------------------------
    // RETURNS = <string> media content or <false> if media is empty
    // ----------------------------------------
        $file = array_shift(explode(";", $val));
        if(count(explode("://", $val)) == 2 || ($file != "" && !is_dir($file) && file_exists($file))) {
            return $val;
        }
        else {
            return false;
        }
    };
    // ====== TEST if MEDIA data is filled / begin ======

// ===================================================
//                   URL pattern
// ===================================================

    function getLinkData($link, $pattern, $path_key = "@path") {
    // ----------------------------------------
    // $link = <string> Website URL link
    // $pattern = <string> Website link pattern, eg: @language/@path/
    // $path_key = <string> Document path key in pattern, default: "@path"
    // ----------------------------------------
    // RETURNS: <array> Read link data: [ "key" => "value", ... ]
    // ----------------------------------------
        $keys = array_flip(array_filter(explode("/", $pattern)));
        $values = array_filter(explode("/", $link));
        $found = [];
        $used = [];
        while(count($values) < count($keys)) { $values[] = ""; }; // Missing values
        // Get non path data
        foreach(array_keys($keys) as $key) {
            if($key != $path_key) {
                $i = $keys[$key];
                if($i > $keys[$path_key]) {
                    $i = count($values) - count($keys) + $i;
                }
                $found[$key] = $values[$i];
                $used[] = $i;
            }
        }
        // Add path
        if(isset($keys[$path_key])) {
            $i = $keys[$path_key];
            $path = [];
            while(!in_array($i, $used) && $i < count($values)) {
                $path[] = $values[$i++];
            };
            $found[$path_key] = join("/", $path);
        }
        // Sort, verify & output
        foreach(array_keys($keys) as $key) { $keys[$key] = $found[$key]; }
        if(join("/", $keys) == join("/", $values)) {
            return $keys;
        }
    };

    function makeLink($link_data, $pattern) {
    // ----------------------------------------
    // $link_data = <array> Read link data: [ "key" => "value", ... ]
    // $pattern = <string> Website link pattern, eg: @language/@path/
    // ----------------------------------------
    // RETURNS: <string> Link maching the pattern
    // ----------------------------------------
        if(!isset($link_data["@language"])) { $link_data["@language"] = $_SESSION["lang"]; };
        foreach(array_keys($link_data) as $key) {
            $pattern = preg_replace("/".$key."/", $link_data[$key], $pattern);
        };
        return $pattern;
    };

// ===================================================
//                      Sitemap
// ===================================================

   // ====== Make Sitemap Item data / begin ======
    function makeSitemapItem($path, $tags, $force_draft = false) {
    // -------------------------------------------
    // $path = <string> XML Document path
    // -------------------------------------------
    // RETURNS: <array> Item data
    // -------------------------------------------
        if(!is_array($tags)) { $tags = []; };
        if($xml = loadXml($path, "draft", $force_draft)) {
            $data = [];
            $keys = array_merge( [ "header title" ], $tags );
            foreach($_SESSION["languages_list"] as $language) {
                foreach($keys as $key) {
                    $tag_words = explode(" ", $key);
                    $section = $xml;
                    foreach($tag_words as $word) {
                        $section = $section[$word][0];
                    }
                    $type = $section["type"][0];
                    $tag = end($tag_words); // the last tag word as key
                    $val = readXml($xml, $key, $language);
                    if($val) {
                        if(in_array($type, [ "text", "textarea" ])) { $data[$tag][$language] = $val; }
                        else { $data[$tag] = $val; }
                    }
                }
            }
            // Status
            if(!file_exists("$path")) { $data["status"] = "unpublished"; }
            elseif(file_exists("$path.draft")) { $data["status"] = "edited"; }
            else { $data["status"] = "published"; }
            // Output
            return $data;
        }
    };
    // ====== Make Sitemap Item data / end ======

    // ====== Make Sitemap (reqursive) / begin ======
    function makeSitemap($root, $folder = "", $tags, $order = false, $force_draft = false) {
    // -------------------------------------------
    // $root = <string> PHP root
    // $folder = <string> Pages folder path
    // $order = <array> Navigation XML array (optional)
    // -------------------------------------------
    // RETURNS: <array> Full site map tree divided into "inside" and "outside" groups
    // -------------------------------------------
        $sitemap = [];
        $done = [];
        if($folder == "") { $folder_path = ""; }
        else { $folder_path = $folder."/"; };
        // Get order items from file
        $order_items = [];
        if(is_array($order)) { $order_items = $order["multi_page"]; }
        elseif($order_xml = loadXml($root.$folder_path.".order")) { $order_items = $order_xml["multi_item"]; };
        // Add all directory
        $order_items = array_merge($order_items, listDir($root.$folder, "xml,draft"));
        // Check every item
        foreach($order_items as $item) {
            if(is_array($item)) { // Item from file
                $location = "inside";
                if($path = readXml($item, "path")) { $href = $folder_path.getFilename($path); } // .order
                else { $href = $folder_path.readXml($item, "href"); }; // navigation.xml
                $data = [];
                $keys = [ "active", "template" ];
                foreach($keys as $key) {
                    if($val = readXml($item, $key)) {
                        $data[$key] = $val;
                    }
                }
            }
            else { // Item from directory
                if(is_array($order)) { $location = "outside"; }
                else { $location = "inside"; };
                $href = $folder_path.getFilename($item);
                $data = [];
            };
            // Add data to map
            if(!in_array($href, $done)) { // Accept new href only
                $data = array_merge($data, makeSitemapItem($root."$href.xml", $tags, $force_draft));
                if(is_dir($root.$href)) {
                    $subpages_sitemap = makeSitemap($root, $href, $tags, false, $force_draft);
                    if($subpages_sitemap && isset($subpages_sitemap["inside"])) {
                        $data["subpages"] = $subpages_sitemap["inside"];
                    }
                }
                if(count($data) > 0) {
                    $sitemap[$location][$href] = $data;
                    $done[] = $href; // to avoid duplications
                }
            };
        };
        // Output
        return $sitemap;
    };
    // ====== Make Sitemap / end ======

    // ====== Launch Sitemap / begin ======
    function launchSitemap($json_path, $pages_root, $tags, $navigation) {
    // -------------------------------------------
    // $json_path = <string> JSON file path (relative to root)
    // $root = <string> PHP root
    // $navigation = <array> Navigation XML array
    // -------------------------------------------
    // RETURNS: <array> Full site map tree divided into "inside" and "outside" groups
    // -------------------------------------------
        if($pages_root != "" && substr($pages_root, (strlen($pages_root) - 1)) != "/") { $pages_root = $pages_root."/"; }
        if((isset($_SESSION["published_flag"]) && $_SESSION["published_flag"] = true) || !file_exists($json_path)) {
            $sitemap = makeSitemap($pages_root, "", $tags, $navigation);
            $json = arrayToJson($sitemap);
            safeSave($root.$json_path, $json);
        }
        else {
            $json = file_get_contents($json_path);
            $sitemap = json_decode($json, true);
        }
        return $sitemap;
    };
    // ====== Launch Sitemap / end ======

    // ====== Get Sitmap Pages Index (reqursive) / begin ======
    function sitemapIndex($map, $language = false) {
    // -------------------------------------------
    // $map = <array> Sitemap tree array
    // -------------------------------------------
    // RETURNS: <array> Index of all site pages & subpages
    // -------------------------------------------
        if(!$language) { $language = $_SESSION["lang"]; };
        if(!$language) { $language = $_SESSION["admin_lang"]; };
        $index = [];
        foreach(array_keys($map) as $href) {
            $data = $map[$href];
            $subpages = [];
            foreach(array_keys($data) as $key) {
                //$val = $data[$key][$language];
                if($key == "subpages") {
                    $subpages = sitemapIndex($data[$key]);
                    $data[$key] = join(";", array_keys($data[$key]));
                }
                else {
                    if(is_array($data[$key])) {
                        $data[$key] = $data[$key][$language];
                    }
                }
            }
            $data["href"] = $href; // Backward compatibility
            $index[$href] = $data;
            if(count($subpages) > 0) {
                $index = array_merge($index, $subpages);
            }
        }
        return $index;
    };
    // ====== Get Sitmap Pages Index (reqursive) / end ======

// ===================================================
//                Debuggging Tools
// ===================================================

    // ====== Array Listing / begin ======
    function arrayList($array, $depth = 0) {
    // ----------------------------------------
    // $array = <array>
    // $depth = <integer> display array level depth (optional)
    // ----------------------------------------
    // Display all array contents (with echo) for debugging
    // ----------------------------------------
        if(!is_numeric($depth)) { $depth = 0; };
        foreach(array_keys($array) as $k) {
            $n = 0;
            while($n++ < $depth) { echo "> "; };
            $val = $array[$k];
            if(is_array($val)) {
                echo "[$k] ...<br>\n";
                arrayList($val, ($depth + 1));
            }
            else {
                echo "[$k] : '$val'<br>\n";
            };
        };
    }; // ====== Array Listing / end ======

    // ====== xml list in html / begin ======
    function xmlList($xml) {
    // ----------------------------------------
    // $xml = <string> XML contents
    // ----------------------------------------
    // Shows XML contents
    // ----------------------------------------
        $xml = str_replace("<", "&#60;", $xml);
        $xml = str_replace(">", "&#62;", $xml);
        $xml = str_replace("\t", "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;", $xml);
        foreach(explode("\n", $xml) as $txt) { echo "$txt<br>"; };
    }; // ====== xml list in html / end ======

?>